% Group fonctions in GNU Prolog
% Author: Guillaume Arnaud
% Date of creation : 12/05/2003
% last change : 26/05/2003

% C gestion of array
% a permutation is represented by the list [a1,a2,...,an] where ai is the position
% image of the ith element of the list.

% Multiplication
% mult(M1,M2,M3,Rep): M1xM2=M3 (Mi a permutation)

mult([],_,[],1):-!.
mult([X1|R1],M2,[X3|R3],1):-
	nth(X1,M2,X3),
	mult(R1,M2,R3,1).

mult((M1Var,M1Val),(M2Var,M2Val),(M3Var,M3Val),2):-
	mult(M1Var,M2Var,M3Var,1),
	liste_inc(M1Val,M1Valinc),
	liste_inc(M2Val,M2Valinc),
	mult(M1Valinc,M2Valinc,M3Valinc,1),
	liste_dec(M3Valinc,M3Val).

liste_inc([],[]).
liste_inc([X|R],[Xp1|R2]):-
	Xp1 is X+1,
	liste_inc(R,R2).
liste_dec([],[]).
liste_dec([X|R],[Xm1|R2]):-
	Xm1 is X-1,
	liste_dec(R,R2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simple Dimino's Algorithm : generate a group with generators.

% dimino(+S,+Rep): S generators, Rep type of representation 

dimino(S,FinalOrder,Rep):-
	nth(1,S,S1), % first symmetry generator
	length(S,T), % T is number of symmetry generators
	
	g_read(numsym_cour,NumSym), % last symmetry number
	id(Id,Rep,S1), % Id is the neutral element
	init_dimino(S1,S1,NumSym,Id,NumSymNew,Rep),!,
	inductive_levels(S,NumSymNew,2,T,FinalOrder,Rep),!.

id(Id,1,S1):-
	length(S1,SizePerm),
	liste_num(Id,1,SizePerm).
id((IdVar,IdVal),2,(LIVar,LVal)):-
	length(LIVar,SizeVar),
	length(LVal,SizeVal),
	EndVal is SizeVal-1,
	liste_num(IdVar,1,SizeVar),
	liste_num(IdVal,0,EndVal).


% init_dimino(+S1,+G,+NumLastSym,+Id,-Order,+Rep) :
% G is nth compsition of S1 until that G=Id
% Order is the new last symmetry number
init_dimino(_,Id,Order,Id,Order,_):-!.
init_dimino(S1,G,NumSym,Id,Order,Rep):-
	NumSymp1 is NumSym+1,
	set_array_pl(NumSymp1,G,Rep),
	mult(G,S1,G2,Rep),
	init_dimino(S1,G2,NumSymp1,Id,Order,Rep).


% inductive_levels(+S,+NumSym,+I,+T,-FinalOrder,+Rep)
inductive_levels(_,FinalOrder,I,T,FinalOrder,_):-I>T,!.
inductive_levels(S,NumSym,I,T,FinalOrder,Rep):-!,
	nth(I,S,SCour),
	length(Si,I),
	append(Si,_,S), % set of i first symmetries
	(membervar(SCour,NumSym,Rep)	->
					NewNumSym=NumSym
					;
					PreviousOrder is NumSym,
					NumSymp1 is NumSym+1,
					set_array_pl(NumSymp1,SCour,Rep),
					g_read(numsym_cour,FirstNumSym),
					CosetStart is FirstNumSym+1,
					% coset of SCour with the previous
					% symmetries found
					coset(SCour,CosetStart,PreviousOrder,NumSymp1,NumSymCoset1,Rep),
					SizeCoset is PreviousOrder-FirstNumSym,
					% coset representative's position
					Rep_pos is PreviousOrder+1,		
					% all other coset
					rest_coset(Si,Si,Rep_pos,NumSymCoset1,SizeCoset,PreviousOrder,RestCosetOrder,Rep),
					NewNumSym=RestCosetOrder
	),
	Ip1 is I+1,
	inductive_levels(S,NewNumSym,Ip1,T,FinalOrder,Rep).
	

% coset(+SCour,+CosetStart,+CosetEnd,+NumSym,-NumSymCoset,+Rep)
% compute the coset (SCour x CosetStart/End)
% out: last number symmetry
coset(_,I,CosetEnd,NumSymCoset,NumSymCoset,_):-
	I>CosetEnd,!.
coset(SCour,I,CosetEnd,NumSym,NumSymCoset,Rep):-!,
	NumSymp1 is NumSym+1,
	get_array_pl(I,Elt,Rep),
	mult(SCour,Elt,CosetElt,Rep),
	set_array_pl(NumSymp1,CosetElt,Rep),
	Ip1 is I+1,
	coset(SCour,Ip1,CosetEnd,NumSymp1,NumSymCoset,Rep).


% rest_coset(+Si,+Si,+Rep_pos,+NumSym,+SizeCoset,+PreviousOrder,-FinalOrder,+Rep)
% find all cosets (Si's element x last generated group)
rest_coset(_,_,Rep_pos,FinalOrder,_,_,FinalOrder,_):-
	Rep_pos>FinalOrder,!.
rest_coset([],Si,Rep_pos,NumSym,SizeCoset,PreviousOrder,FinalOrder,Rep):-!,
	NewRep_pos is Rep_pos+SizeCoset,
	rest_coset(Si,Si,NewRep_pos,NumSym,SizeCoset,PreviousOrder,FinalOrder,Rep).
rest_coset([SElt|Si],S,Rep_pos,NumSym,SizeCoset,PreviousOrder,FinalOrder,Rep):-!,
	get_array_pl(Rep_pos,Eltrp,Rep),
	mult(Eltrp,SElt,Elt,Rep),
	((id(Elt,2,Elt);membervar(Elt,NumSym,Rep))->
				rest_coset(Si,S,Rep_pos,NumSym,SizeCoset,PreviousOrder,FinalOrder,Rep)
				;
				% Add coset
				NumSymp1 is NumSym+1,
				set_array_pl(NumSymp1,Elt,Rep),
				g_read(numsym_cour,FirstNumSym),
				CosetStart is FirstNumSym+1,
				coset(Elt,CosetStart,PreviousOrder,NumSymp1,NumSymCoset,Rep),
				rest_coset(Si,S,Rep_pos,NumSymCoset,SizeCoset,PreviousOrder,FinalOrder,Rep)				
				).

membervar(Elt,NumSym,Rep):-
	g_read(numsym_cour,NumSymCour),
	NumSymCourp1 is NumSymCour+1,
	membervar_aux(Elt,NumSym,NumSymCourp1,Rep).
membervar_aux(_,NumSym,I,_):-
	I>NumSym,!,fail.
membervar_aux(Elt,_,I,Rep):-
	get_array_pl(I,Elt,Rep),!.
membervar_aux(Elt,NumSym,I,Rep):-
	Ip1 is I+1,
	membervar_aux(Elt,NumSym,Ip1,Rep).


% Array gestion
set_array_pl(NumSym,Element,Rep):-
	set_array_elt_pl(NumSym,Element,1,Rep).
	
set_array_elt_pl(_,[],_,1):-!.
set_array_elt_pl(NumSym,[XSym|Element],X,1):-!,
	set_grp_point_c(NumSym,X,XSym),
	Xp1 is X+1,
	set_array_elt_pl(NumSym,Element,Xp1,1).
set_array_elt_pl(NumSym,([XSym|LVar],LVal),X,2):-!,
	set_grp_var_c(NumSym,X,XSym),
	Xp1 is X+1,
	set_array_elt_pl(NumSym,(LVar,LVal),Xp1,2).
set_array_elt_pl(NumSym,([],LVal),_,2):-!,
	set_array_elt_pl(NumSym,LVal,0,2).
set_array_elt_pl(_,[],_,2):-!.
set_array_elt_pl(NumSym,[XSym|Element],X,2):-!,
	set_grp_val_c(NumSym,X,XSym),
	Xp1 is X+1,
	set_array_elt_pl(NumSym,Element,Xp1,2).
	
get_array_pl(NumSym,Element,1):-
	g_read(nbpoint,NbPoint),
	length(Element,NbPoint),
	get_array_elt_pl(NumSym,Element,1,1).
get_array_pl(NumSym,(LVar,LVal),2):-
	g_read(nbvar,NbVar),
	length(LVar,NbVar),
	g_read(nbval,NbVal),
	length(LVal,NbVal),
	get_array_elt_pl(NumSym,(LVar,LVal),1,2).
get_array_elt_pl(_,[],_,1):-!.
get_array_elt_pl(_,[],_,2):-!.
get_array_elt_pl(NumSym,[XSym|Element],X,1):-!,
	get_grp_point_c(NumSym,X,XSym),
	Xp1 is X+1,
	get_array_elt_pl(NumSym,Element,Xp1,1).
get_array_elt_pl(NumSym,([XSym|LVar],LVal),X,2):-!,
	get_grp_var_c(NumSym,X,XSym),
	Xp1 is X+1,
	get_array_elt_pl(NumSym,(LVar,LVal),Xp1,2).
get_array_elt_pl(NumSym,([],LVal),_,2):-!,
	get_array_elt_pl(NumSym,LVal,0,2).
get_array_elt_pl(NumSym,[XSym|LVal],X,2):-!,
	get_grp_val_c(NumSym,X,XSym),
	Xp1 is X+1,
	get_array_elt_pl(NumSym,LVal,Xp1,2).



